@@ -1,27 +1,29 @@
 # The R Commander and command logger
-
-# laste modified 1 december 2006 by Adrian Dusa
-# based on the latest modifications since 6 October 06 by J. Fox
+# last modified 07 October 2007 by Adrian Dusa
+#   based on the latest modifications since 26 July 2007 by J. Fox
 #   slight changes 12 Aug 04 by Ph. Grosjean 
+#   changes 21 June 2007 by Erich Neuwirth for Excel support (marked EN)
 
 Commander <- function(){
+    require("MASS")
+    RcmdrVersion <- "1.3-5"
     # the following test suggested by Richard Heiberger
     if ("RcmdrEnv" %in% search() &&
         exists("commanderWindow", "RcmdrEnv") &&
         !is.null(get("commanderWindow", "RcmdrEnv"))) {
-      warning("The QCAGUI is already open.")
+      warning("QCAGUI is already open.")
       return(invisible(NULL))
     }
-    RcmdrVersion <- "1.2-4"
     if (is.SciViews()) return(invisible(svCommander(Version=RcmdrVersion))) # +PhG
     setOption <- function(option, default, global=TRUE) {
-        opt <- if (is.null(current[[option]])) default else current[[option]]
+        opt <- if (is.null(current[option][[1]])) default else current[option][[1]]
         if (global) putRcmdr(option, opt)
         else opt
         }
     current <- options("Rcmdr")[[1]]
-    etc <- setOption("etc", file.path(.path.package(package="QCAGUI")[1], "etc"), global=FALSE)
-    etcMenus <- setOption("etcMenus", etc, global=FALSE)
+    etc <- setOption("etc", file.path(.path.package(package="QCAGUI")[1], "etc"))
+    etcMenus <- setOption("etcMenus", etc)
+    putRcmdr("etcMenus", etcMenus)
     onCopy <- function(){
         focused <- tkfocus()
         if ((tclvalue(focused) != LogWindow()$ID) && (tclvalue(focused) != OutputWindow()$ID))
@@ -173,7 +175,7 @@ Commander <- function(){
     placement <- setOption("placement", "-40+20", global=FALSE)
     tkwm.geometry(.commander, placement)
     tkwm.title(.commander, gettextRcmdr("Qualitative Comparative Analysis"))
-    tkwm.protocol(.commander, "WM_DELETE_WINDOW", closeCommander)
+    tkwm.protocol(.commander, "WM_DELETE_WINDOW", CloseCommander)
     topMenu <- tkmenu(.commander)
     tkconfigure(.commander, menu=topMenu)
     source.files <- list.files(etc, pattern="\\.[Rr]$")
@@ -181,45 +183,127 @@ Commander <- function(){
         source(file.path(etc, file))
         cat(paste(gettextRcmdr("Sourced:"), file, "\n"))
         }
-    Menus <- read.table(file.path(etcMenus, "QCA-menus.txt"), as.is=TRUE)
+    Plugins <- options()$Rcmdr$plugins
+    allPlugins <- listPlugins(loaded=TRUE)
+    for (plugin in Plugins){
+        if (!require(plugin, character.only=TRUE)){
+            putRcmdr("commanderWindow", NULL) 
+            stop(sprintf(gettextRcmdr("the plug-in package %s is missing"), plugin))
+            }
+        if (!is.element(plugin, allPlugins)){
+            putRcmdr("commanderWindow", NULL) 
+            stop(sprintf(gettextRcmdr("the package %s is not an Rcmdr plug-in"), plugin))
+            }
+        }
+    Menus <- read.table(file.path(etcMenus, "QCA-menus.txt"), colClasses = "character")
+    addMenus <- function(Menus){
+        nms <- c("type", "menuOrItem", "operationOrParent", "label",
+            "commandOrMenu", "activation", "install")
+        names(Menus) <- nms        
+        for (plugin in Plugins) {
+            MenusToAdd <- read.table(file.path(.path.package(package=plugin)[1], "etc/menus.txt"), 
+                colClasses = "character")
+            names(MenusToAdd) <- nms
+            for (i in 1:nrow(MenusToAdd)){
+                line <- MenusToAdd[i,]
+                if (line[1, "type"] == "menu"){
+                    where <- if (line[1, "operationOrParent"] == "topMenu") 0
+                        else which((Menus[, "type"] == "menu") & 
+                            (Menus[, "menuOrItem"] == line[1, "operationOrParent"]))
+                    }
+                else if (line[1, "type"] == "item"){
+                    if (line[1, "operationOrParent"] == "command"){
+                        which <- which((Menus[, "operationOrParent"] == "command") &
+                            (Menus[, "menuOrItem"] == line[1, "menuOrItem"]))
+                        where <- if (length(which) == 0) 
+                            which((Menus[, "type"] == "menu")
+                                & (Menus[, "menuOrItem"] == line[1, "menuOrItem"]))
+                            else max(which)
+                        }
+                    else if (line[1, "operationOrParent"] == "cascade"){
+                        where <- if (line[1, "menuOrItem"] != "topMenu")
+                            max(which((Menus[, "operationOrParent"] == "cascade") &
+                                (Menus[, "menuOrItem"] == line[1, "menuOrItem"])))
+                        else {
+                            max(which((Menus[, "operationOrParent"] == "cascade") &
+                                (Menus[, "menuOrItem"] == "topMenu") &
+                                (Menus[, "commandOrMenu"] != "toolsMenu") &
+                                (Menus[, "commandOrMenu"] != "helpMenu")))
+                            }                        
+                        }
+                    else stop(sprintf(gettextRcmdr('unrecognized operation, "%s", in plugin menu line %i'), 
+                        line[1, "operation"], i))
+                    }
+                else stop(sprintf(gettextRcmdr('unrecognized type, "%s", in plugin menu line %i'),
+                   line[1, "type"], i))
+                Menus <- insertRows(Menus, line, where)
+                }
+            }
+        Menus
+        }
+    Menus <- addMenus(Menus)
     .Menus <- menus <- list()
     menuItems <- 0
     oldMenu <- ncol(Menus) == 6
-    for (m in 1:nrow(Menus)){
-        install <- if (oldMenu) "" else Menus[m, 7]
-        if ((install != "") && (!eval(parse(text=install)))) next  
-        if (Menus[m, 1] == "menu") {
-            position <- 0
-            assign(Menus[m, 2], tkmenu(eval(parse(text=Menus[m, 3])), tearoff=FALSE))
-            menus[[Menus[m, 2]]] <- list(ID=get(Menus[m, 2])$ID, position=0)
-            }
-        else if (Menus[m, 1] == "item") {
-            if (Menus[m, 3] == "command"){
-                position <- position + 1
-                if (Menus[m, 6] == "")
-                    tkadd(eval(parse(text=Menus[m, 2])),"command", label=gettextRcmdr(Menus[m, 4]),
-                        command=eval(parse(text=Menus[m, 5])))
-                else {
-                    tkadd(eval(parse(text=Menus[m, 2])),"command", label=gettextRcmdr(Menus[m, 4]),
-                        command=eval(parse(text=Menus[m, 5])),  state="disabled")
-                    menuItems <- menuItems + 1
-                    menus[[Menus[m, 2]]]$position <- position
-                    .Menus[[menuItems]] <- list(ID=menus[[Menus[m, 2]]]$ID, position=position,
-                        activation=eval(parse(text=paste("function()", Menus[m, 6]))))
+    setOption("suppress.menus", FALSE)
+    if (!getRcmdr("suppress.menus")){
+        for (m in 1:nrow(Menus)){
+            install <- if (oldMenu) "" else Menus[m, 7]
+            if ((install != "") && (!eval(parse(text=install)))) next  
+            if (Menus[m, 1] == "menu") {
+                position <- 0
+                assign(Menus[m, 2], tkmenu(eval(parse(text=Menus[m, 3])), tearoff=FALSE))
+                menus[[Menus[m, 2]]] <- list(ID=get(Menus[m, 2])$ID, position=0)
+                }
+            else if (Menus[m, 1] == "item") {
+                if (Menus[m, 3] == "command"){
+                    position <- position + 1
+                    if (Menus[m, 6] == "")
+                        tkadd(eval(parse(text=Menus[m, 2])),"command", label=gettextRcmdr(Menus[m, 4]),
+                            command=eval(parse(text=Menus[m, 5])))
+                    else {
+                        tkadd(eval(parse(text=Menus[m, 2])),"command", label=gettextRcmdr(Menus[m, 4]),
+                            command=eval(parse(text=Menus[m, 5])),  state="disabled")
+                        menuItems <- menuItems + 1
+                        menus[[Menus[m, 2]]]$position <- position
+                        .Menus[[menuItems]] <- list(ID=menus[[Menus[m, 2]]]$ID, position=position,
+                            activation=eval(parse(text=paste("function()", Menus[m, 6]))))
+                        }
                     }
+                else if (Menus[m, 3] == "cascade")
+                    tkadd(eval(parse(text=Menus[m, 2])),"cascade", label=gettextRcmdr(Menus[m, 4]), 
+                        menu=eval(parse(text=Menus[m, 5])))
+                else stop(paste(gettextRcmdr("menu definition error:"), Menus[m, ], collapse=" "),
+                    domain=NA)
                 }
-            else if (Menus[m, 3] == "cascade")
-                tkadd(eval(parse(text=Menus[m, 2])),"cascade", label=gettextRcmdr(Menus[m, 4]), menu=eval(parse(text=Menus[m, 5])))
             else stop(paste(gettextRcmdr("menu definition error:"), Menus[m, ], collapse=" "),
                 domain=NA)
             }
-        else stop(paste(gettextRcmdr("menu definition error:"), Menus[m, ], collapse=" "),
-            domain=NA)
         }
+    ## added by EN ###############################
+	if (RExcelSupported())
+    	putRExcel(".rexcel.menu.dataframe", Menus)
+    ## end of change ###############################
     putRcmdr("Menus", .Menus)
+    putRcmdr("autoRestart", FALSE)
     activateMenus()
     exceptions <- scan(file.path(etc, "log-exceptions.txt"), what="", quiet=TRUE, comment.char="#")
-    putRcmdr("modelClasses", scan(file.path(etc, "model-classes.txt"), what="", quiet=TRUE, comment.char="#"))
+    ## added by EN ###############################
+   	putRcmdr("Commander.Input.exceptions", exceptions)
+    ## end of change ###############################
+    modelClasses <- scan(file.path(etc, "model-classes.txt"), what="", quiet=TRUE, comment.char="#")
+    for (plugin in Plugins){
+        description <- readLines(file.path(.path.package(package=plugin)[1], "DESCRIPTION"))
+        addExceptions <- description[grep("Log-Exceptions:", description)]
+        addExceptions <- gsub(" ", "", sub("^Log-Exceptions:", "", addExceptions))
+        addExceptions <- unlist(strsplit(addExceptions, ","))
+        addModels <- description[grep("Models:", description)]
+        addModels <- gsub(" ", "", sub("^Models:", "", addModels))
+        addModels <- unlist(strsplit(addModels, ","))
+        if (length(addExceptions) > 0) exceptions <- c(exceptions, addExceptions)
+        if (length(addModels) > 0) modelClasses <- c(modelClasses, addModels)
+        }
+    putRcmdr("modelClasses", modelClasses)
     onEdit <- function(){
         if (activeDataSet() == FALSE) {
             tkfocus(CommanderWindow())
@@ -244,18 +328,19 @@ Commander <- function(){
             paste("showData(", ActiveDataSet(), ", placement='-20+200', font=getRcmdr('logFont'), maxwidth=",
                 log.width, ", maxheight=", view.height, ")", sep="")
             }
-        else paste("invisible(edit(", ActiveDataSet(), "))", sep="")
+        else paste("View(", ActiveDataSet(), ")", sep="")
         logger(command)
         justDoIt(command)
         tkwm.deiconify(CommanderWindow())
         tkfocus(CommanderWindow())
         }
+    # the following function modified 14 July 07 by Erich Neuwirth
     onSubmit <- function(){
         .log <- LogWindow()
         selection <- strsplit(tclvalue(tktag.ranges(.log, "sel")), " ")[[1]]
         if (is.na(selection[1])) {
             tktag.add(.log, "currentLine", "insert linestart", "insert lineend")
-            selection <- strsplit(tclvalue(tktag.ranges(.log, "currentLine")), " ")[[1]]
+            selection <- strsplit(tclvalue(tktag.ranges(.log,"currentLine")), " ")[[1]]
             tktag.delete(.log, "currentLine")
             if (is.na(selection[1])) {
                 Message(message=gettextRcmdr("Nothing is selected."),
@@ -268,30 +353,33 @@ Commander <- function(){
         lines <- strsplit(lines, "\n")[[1]]
         .console.output <- getRcmdr("console.output")
         .output <- OutputWindow()
-        if (!.console.output) tkinsert(.output, "end", "\n")
         iline <- 1
         nlines <- length(lines)
         while (iline <= nlines){
+            while (nchar(lines[iline])==0) iline <- iline + 1
+            if (iline > nlines) break
             current.line <- lines[iline]
             if (.console.output) cat(paste("\nQCA> ", current.line,"\n", sep=""))
             else{
-                tkinsert(.output, "end", paste("> ", current.line,"\n", sep=""))
+                tkinsert(.output, "end", paste("\n> ", current.line,"\n", sep="")) ### end of changed
                 tktag.add(.output, "currentLine", "end - 2 lines linestart", "end - 2 lines lineend")
                 tktag.configure(.output, "currentLine", foreground=getRcmdr("command.text.color"))
                 }
             jline <- iline + 1
             while (jline <= nlines){
-                if (length(grep("^[\\ \t]", lines[jline])) == 0) break
-                if (.console.output) cat(paste("QCA+ ", lines[jline],"\n", sep=""))
+                   if (class(try(parse(text=current.line),silent=TRUE))!="try-error") break
+                if (.console.output)cat(paste("QCA+ ", lines[jline],"\n", sep=""))
                 else{
                     tkinsert(.output, "end", paste("+ ", lines[jline],"\n", sep=""))
                     tktag.add(.output, "currentLine", "end - 2 lines linestart", "end - 2 lines lineend")
                     tktag.configure(.output, "currentLine", foreground=getRcmdr("command.text.color"))
                     }
-                current.line <- paste(current.line, lines[jline])
+                current.line <- paste(current.line, lines[jline],sep="\n")
                 jline <- jline + 1
                 iline <- iline + 1
                 }
+            if (!(is.null(current.line) || is.na(current.line))){
+
             if (length(grep("<-", current.line)) > 0){
                 justDoIt(current.line)
                 }
@@ -304,10 +392,11 @@ Commander <- function(){
                 justDoIt(current.line)
                 }
             else doItAndPrint(current.line, log=FALSE)
-            iline <- iline + 1
             }
+            iline <- iline + 1
         tkyview.moveto(.output, 1)
         }
+    }
     contextMenuLog <- function(){
         .log <- LogWindow()
         contextMenu <- tkmenu(tkmenu(.log), tearoff=FALSE)
@@ -381,11 +470,13 @@ Commander <- function(){
         relief="groove", command=selectActiveModel))
     show.edit.button <- options("Rcmdr")[[1]]$show.edit.button
     show.edit.button <- if (is.null(show.edit.button)) TRUE else show.edit.button
-    tkgrid(tklabel(controlsFrame, bitmap=paste("@", file.path(etc, "QCA.xbm"), sep=""), fg="red"),
-        tklabel(controlsFrame, text=gettextRcmdr("Data set:")), getRcmdr("dataSetLabel"),
-        tklabel(controlsFrame, text="  "), if(show.edit.button) editButton, viewButton, sticky="w")
-        #tklabel(controlsFrame, text=gettextRcmdr("    Model: ")), getRcmdr("modelLabel"), sticky="w")
-    tkgrid(controlsFrame, sticky="w")
+    if (!getRcmdr("suppress.menus")){
+        tkgrid(tklabel(controlsFrame, bitmap=paste("@", file.path(etc, "QCA.xbm"), sep=""), fg="red"),
+            tklabel(controlsFrame, text=gettextRcmdr("Data set:")), getRcmdr("dataSetLabel"),
+            tklabel(controlsFrame, text="  "), if(show.edit.button) editButton, viewButton, sticky="w")
+            # tklabel(controlsFrame, text=gettextRcmdr("    Model: ")), getRcmdr("modelLabel"), sticky="w")
+        tkgrid(controlsFrame, sticky="w")
+        }
     .log.commands <-  getRcmdr("log.commands")
     .console.output <- getRcmdr("console.output")
     if (.log.commands) tkgrid(tklabel(logFrame, text=gettextRcmdr("Script Window"), fg="blue"),
@@ -456,25 +547,36 @@ Commander <- function(){
     Message(paste(gettextRcmdr("QCAGUI based on R Commander Version "), getRcmdr("RcmdrVersion"), ": ", date(), sep=""))
     }
 
-logger <- function(command){
-    if (is.SciViews()) return(svlogger(command))    # +PhG
-    .log <- LogWindow()
-    .output <- OutputWindow()
-    if (getRcmdr("log.commands")) {
-        tkinsert(.log, "end", paste(command,"\n", sep=""))
-        tkyview.moveto(.log, 1)
-        }
-    lines <- strsplit(command, "\n")[[1]]
-    tkinsert(.output, "end", "\n")
-    if (getRcmdr("console.output")) for (line in lines) cat(paste("\nQCA>", line, "\n"))
-    else {
-        for (line in lines) tkinsert(.output, "end", paste("> ", command,"\n", sep=""))
-        tktag.add(.output, "currentLine", "end - 2 lines linestart", "end - 2 lines lineend")
-        tktag.configure(.output, "currentLine", foreground=getRcmdr("command.text.color"))
-        tkyview.moveto(.output, 1)
-        }
-    command
-    }
+# the following function modified 24 July 07 by Richard Heiberger
+#  and subsequently by J. Fox 26 July 07
+
+logger <- function(command){ 
+   if (is.SciViews()) return(svlogger(command))    # +PhG 
+   .log <- LogWindow() 
+   .output <- OutputWindow() 
+   if (getRcmdr("log.commands")) { 
+       tkinsert(.log, "end", paste(command,"\n", sep="")) 
+       tkyview.moveto(.log, 1) 
+       } 
+   lines <- strsplit(command, "\n")[[1]] 
+   tkinsert(.output, "end", "\n") 
+   if (getRcmdr("console.output")) { 
+     for (line in seq(along=lines)) { 
+       prompt <- ifelse (line==1, "\nQCA>", "\nQCA+") 
+       cat(paste(prompt, lines[line], "\n")) 
+        } 
+     } 
+   else { 
+     for (line in  seq(along=lines)) { 
+       prompt <- ifelse(line==1, "> ", "+ ") 
+       tkinsert(.output, "end", paste(prompt, lines[line], "\n", sep=""))
+       tktag.add(.output, "currentLine", "end - 2 lines linestart", "end - 2 lines lineend") 
+       tktag.configure(.output, "currentLine", foreground=getRcmdr("command.text.color")) 
+       tkyview.moveto(.output, 1)  
+       } 
+     } 
+   command 
+   } 
 
 justDoIt <- function(command) {
     Message()
@@ -531,7 +633,8 @@ doItAndPrint <- function(command, log=TRUE) {
         }
     if (isS4object(result)) show(result) else print(result)
     .Output <- readLines(output.connection)
-    if (.Output[length(.Output)] == "NULL") .Output <- .Output[-length(.Output)] # suppress "NULL" line at end of output
+    if (length(.Output) > 0 && .Output[length(.Output)] == "NULL") 
+        .Output <- .Output[-length(.Output)] # suppress "NULL" line at end of output
     if (length(.Output) != 0) {  # is there output to print?
         if (.console.output) {
             out <- .Output
@@ -544,6 +647,10 @@ doItAndPrint <- function(command, log=TRUE) {
             }
         }
     else if (.console.output) sink(type="output")
+    ###### added by EN  ######################
+	if (RExcelSupported())
+    	putRExcel(".rexcel.last.output",.Output)
+    ###### end of change  #####################
     # errors already intercepted, display any warnings
     checkWarnings(readLines(messages.connection))
     result
@@ -606,6 +713,10 @@ Message <- function(message, type=c("note", "error", "warning")){
         }
     putRcmdr("last.message", type)
     message <- paste(prefix, ": ", message, sep="")
+    ######### added by EN #####################
+   	if (RExcelSupported())
+    	putRExcel(".rexcel.last.message",message)
+    ######### end of change ###############
     lines <- strsplit(message, "\n")[[1]]
     for (line in lines){
         tagName <- messageTag()
